#include <sys/time.h>
#include <unistd.h>
#include <malloc.h>
#include <stdio.h>
#include "levelcurve.h"
#include "tasklist.h"
#include "list.h"
#include "points.h"


#define CLOSED             0x01
#define ISREAL             0x02
#define LIMIT_2_TRIANGLES  0x04
#define DIRECT_CLOSED      0x08
#define REVERSE_CLOSED     0x10

#define COSPI_3		   0.5
#define SINPI_3		   0.86602540378444

/*===================== Warning===============================
  Same constants are defined in userfunc.c - DO NOT MODIFY
**************************************************************/

#define Zi_CONV            0x01
#define Ze_CONV            0x02
#define Zl_INCL            0x04
#define Zl_CONV            0x08

/********************* End Warning *************************/


static int SocketHandle=-1;

struct Segment
{
  long xi,yi,xe,ye;
  char Flag;
  Complex z;
  double Sigma;
};

struct LevelCurve
{
  int Id;
  int Priority;
  char Method;
  int MinP,MaxP; /* MinP<=Processor<=MaxP*/
  double Tau;
  double Sigma,SigmaPrec,SplitPrec;
  double CreatedTime;
  double LastModified;
  long SigmaMaxIt;
  char OrbitFlags;
  void *Orbit; /* Liste des points */
  void *TaskList;
  struct LevelCurve *Next;
};


struct Task
{
  int Tag;
  void *Segment;
  struct MasterMsg Msg; 
};

static struct LevelCurve *LevelList=NULL;
static char s[1000];

static double tic()
{
  static double ref=-1.0;
  if (ref==-1.0)
    {
      ref=time(0);
      return 0.0;
    }
  return time(0)-ref;
}


static void SendSock(int sock,char *s)
{
  if (sock==-1)
    return;
  write(sock,s,strlen(s));
}


int CreateLevelCurve(Complex zi,double Tau,double Sigma,
		     double SigmaPrec,long SigmaMaxIt,
		     double SplitPrec,char IsReal,char Method,double angle)
{
  struct LevelCurve *L;
  struct Task T;
  static int Id=0;

  if ((L=(struct LevelCurve *)malloc(sizeof(struct LevelCurve)))==NULL)
    return -1;
  L->Method=Method;
  L->Id=++Id;
  if (IsReal) /* Si la matrice est reelle, le pseudospectre est symetrique */
    L->OrbitFlags=ISREAL;
  else
    L->OrbitFlags=0;
  L->MinP=-1; /* Tous les processeurs */
  L->TaskList=CreateTaskList(50);
  L->CreatedTime=tic();
  /* Donnees concernant la ligne de niveau */
  L->Tau=Tau;
  L->Sigma=Sigma;
  L->SigmaPrec=SigmaPrec;
  L->SigmaMaxIt=SigmaMaxIt;
  L->SplitPrec=SplitPrec;
  L->Orbit=CreateEmptyList();
  L->Next=LevelList;
  LevelList=L;
  T.Tag=STARTUP_TAG;
  T.Msg.z1=zi;
  T.Msg.z2.r=Tau;  
  T.Msg.z2.i=angle;
  T.Msg.Sigma=Sigma;
  T.Msg.SigmaPrec=SigmaPrec;
  T.Msg.SigmaMaxIt=SigmaMaxIt;
  T.Msg.Method=Method;
  AddHighPriorityTask(&T,sizeof(struct Task),L->TaskList);
  SetLevelCurvePriority(L->Id,0);
  return L->Id;
}

int DestroyLevelCurve(int Id)
{
  struct LevelCurve *p,*q;
  struct Task T;
  p=LevelList;
  while(p!=NULL && p->Id!=Id)
    {
      q=p;
      p=p->Next;
    }
  if (p==NULL)
    return 0;
  if (p==LevelList)
    LevelList=LevelList->Next;
  else
    q->Next=p->Next;
  /* Liberer toutes les taches de calcul */
  while(GetTaskFromList(&T,sizeof(T),p->TaskList));
  DestroyList(p->Orbit);
  free(p);
  return 1;
}

static int PrintCell(void *cell,void *fi)
{
  struct Segment *s=(struct Segment *)cell;
  FILE *f=(FILE *) fi;

  fprintf(f,"%5ld %5ld %5ld %5ld %x %.5e %.5e %e\n",
	 s->xi,s->yi,s->xe,s->ye,s->Flag,
	 s->z.r,s->z.i,s->Sigma);
  return 0;
}

static void Print(struct LevelCurve *L,FILE *f)
{
 if (L==NULL || f==NULL)
  return;
 fprintf(f,"%%Sigma=%e\n",L->Sigma);
 fprintf(f,"%%SigmaPrec=%e\n",L->SigmaPrec);
 fprintf(f,"%%LanczosMaxIt=%ld\n",L->SigmaMaxIt);
 fprintf(f,"%%SplitPrec=%e\n",L->SplitPrec);
 fprintf(f,"%%Created Time=%fs\n",L->CreatedTime);
 fprintf(f,"%%Last Modified=%fs\n",L->LastModified);
 fprintf(f,"%%   i  ,  xi ,  yi ,  xe ,  ye ,   z.r  ,   z.i  ,   Sigma  \n");
 ForAllData(L->Orbit,PrintCell,(void *)f);
}

void PrintLevelCurve(int Id)
{
 struct LevelCurve *L=LevelList;
 while (L!=NULL && L->Id!=Id)
   L=L->Next;
 if (L==NULL)
   return;
 Print((struct LevelCurve *)L,stdout);
 fflush(stdout);
}

int SaveLevelCurve(int Id,char *filename)
{
  FILE *f;
 struct LevelCurve *L=LevelList;
 while (L!=NULL && L->Id!=Id)
   L=L->Next;
 if (L==NULL)
   return -1;
  if ((f=fopen(filename,"wt"))==NULL)
    return -1;
  Print((struct LevelCurve *)L,f);
  fclose(f);
  return 1;
}


/* Retourne un message a envoyer au processeur Processor */
int GetMasterMessage(int Processor,struct MasterMsg *Msg,
		     int *Tag,struct ProcInfo *Proc)
{
  struct LevelCurve *L=LevelList;
  struct Task T;
  while (L)
    {
      /* Verifier si l'utilisation de la ligne de niveau est authorizee
	 pour Processor */
      if (L->MinP==-1 || (L->MinP<=Processor && Processor <=L->MaxP))
	{
	  if (GetTaskFromList(&T,sizeof(T),L->TaskList))
	    {
	      *Msg=T.Msg;
	      *Tag=T.Tag;
	      Proc->LevelCurve=L;
	      Proc->Segment=T.Segment;
	      return 1;
	    }
	}
      /* Passer a la ligne suivante */
      L=L->Next;
    }
  return 0;
}

static int SameCell(long x0i,long y0i,
		    long x0e,long y0e,
		    long x0u,long y0u,
		    long x1i,long y1i,
		    long x1e,long y1e,
		    long x1u,long y1u)
{
  /* EUI==EIU */
  if (x0e==x1e && 
      x0u==x1i &&
      x0i==x1u &&
      y0e==y1e && 
      y0u==y1i &&
      y0i==y1u)
    return 1;
  /* UIE==EIU*/
  if (x0u==x1e &&
      x0i==x1i &&
      x0e==x1u &&
      y0u==y1e &&
      y0i==y1i &&
      y0e==y1u)
    return 1;
 return 0;
}    

static void PredictInverse(struct LevelCurve *Lev)
{
 long x,y;
 int i;
 struct Task T;
 struct Segment S,*Si,*Sd;
 int Rot[6][2]={{1,0},{0,1},{-1,1},{-1,0},{0,-1},{1,-1}};
 double f;
 struct LevelCurve *tmp;

 if (Lev->OrbitFlags & REVERSE_CLOSED)
   {
     return;
   }
 Si=GetLastDataAdress(Lev->Orbit);
 x=Si->xe-Si->xi;
 y=Si->ye-Si->yi;
 for(i=0;i<6;i++)
  if (Rot[i][0]==x && Rot[i][1]==y)
   break;
 i=(i+5)%6;
 x=Si->xi+Rot[i][0];
 y=Si->yi+Rot[i][1];
 if (y<0 && (Lev->OrbitFlags & ISREAL))
 {
   Lev->OrbitFlags |=REVERSE_CLOSED;
   return;
 }
 tmp=LevelList;
 while(tmp)
   {
     if (tmp->Tau==Lev->Tau && 
	 tmp->Sigma == Lev->Sigma &&
	 tmp->SigmaPrec==Lev->SigmaPrec)
       {
	 if ((tmp->OrbitFlags & DIRECT_CLOSED)!=0)
	   {
	     tmp=tmp->Next;
	     continue;
	   }
	 Sd=GetFirstDataAdress(tmp->Orbit);
	 if (Sd && SameCell(Si->xi,Si->yi,
			    Si->xe,Si->ye,
			    x,y,
			    Sd->xi,Sd->yi,
			    Sd->xe,Sd->ye,
			    (long) Sd->z.r, (long) Sd->z.i))
	 {
	     tmp->OrbitFlags |=DIRECT_CLOSED;
	     Lev->OrbitFlags |= REVERSE_CLOSED;
	     return;
	   }
       }
     tmp=tmp->Next;
   }
 S.z.r=x;
 S.z.i=y;
 S.xi=Si->xi;
 S.yi=Si->yi;
 S.xe=Si->xe;
 S.ye=Si->ye;
 S.Flag=Si->Flag & (Ze_CONV | Zi_CONV);
 AddLast(&S,sizeof(S),Lev->Orbit);
 /* Check for computed points */
 if (Lookup(x,y,Lev->Tau,&f,Lev->SigmaPrec))
   {
     T.Msg.Sigma=f;
     T.Msg.Method=-1;
   }
 else
 T.Msg.Method=Lev->Method;
 T.Tag=Z1_TAG;
 T.Msg.z1.r=x*Lev->Tau+y*Lev->Tau*COSPI_3;
 T.Msg.z1.i=y*Lev->Tau*SINPI_3;
 T.Msg.SigmaPrec=Lev->SigmaPrec;
 T.Msg.SigmaMaxIt=Lev->SigmaMaxIt;
 AddHighPriorityTask(&T,sizeof(struct Task),Lev->TaskList);
}


static void PredictDirect(struct LevelCurve *Lev)
{
 long x,y;
 int i;
 struct Task T;
 struct Segment S,*Sd,*Si;
 static int Rot[6][2]={{1,0},{0,1},{-1,1},{-1,0},{0,-1},{1,-1}};
 double f;
 struct LevelCurve *tmp;

 if (Lev->OrbitFlags & DIRECT_CLOSED)
   {
     return;
   }
 Sd=GetFirstDataAdress(Lev->Orbit);
 x=Sd->xe-Sd->xi;
 y=Sd->ye-Sd->yi;
 for(i=0;i<6;i++)
  if (Rot[i][0]==x && Rot[i][1]==y)
   break;
 i=(i+1)%6;
 x=Sd->xi+Rot[i][0];
 y=Sd->yi+Rot[i][1];
 if (y<0 && (Lev->OrbitFlags & ISREAL))
 {
   Lev->OrbitFlags |=DIRECT_CLOSED;
   return;
 }
 tmp=LevelList;
 while (tmp)
   {
     if (tmp->Tau==Lev->Tau && 
	 tmp->Sigma==Lev->Sigma &&
	 tmp->SigmaPrec==Lev->SigmaPrec)
       {
	 if ((tmp->OrbitFlags & REVERSE_CLOSED)!=0)
	   {
	     tmp=tmp->Next;
	     continue;
	   }
	 Si=GetLastDataAdress(tmp->Orbit);
	 if (Si && SameCell(Si->xi,Si->yi,
			    Si->xe,Si->ye,
			    (long) Si->z.r,(long) Si->z.i,
			    Sd->xi,Sd->yi,
			    Sd->xe,Sd->ye,
			    x,y))
	   {
	     Lev->OrbitFlags |=DIRECT_CLOSED;
	     tmp->OrbitFlags |=REVERSE_CLOSED;
	     return;
	   }
       }
     tmp=tmp->Next;
   }
 S.z.r=x;
 S.z.i=y;
 S.xi=Sd->xi;
 S.yi=Sd->yi;
 S.xe=Sd->xe;
 S.ye=Sd->ye;
 S.Flag=Sd->Flag & (Ze_CONV | Zi_CONV);
 AddFirst(&S,sizeof(S),Lev->Orbit);
 /* Check for computed points */
 if (Lookup(x,y,Lev->Tau,&f,Lev->SigmaPrec))
   {
     T.Msg.Sigma=f;
     T.Msg.Method=-1;
   }
 else
   T.Msg.Method=Lev->Method;
 T.Tag=Z_TAG;
 T.Msg.z1.r=x*Lev->Tau+y*Lev->Tau*COSPI_3;
 T.Msg.z1.i=y*Lev->Tau*SINPI_3;
 T.Msg.SigmaPrec=Lev->SigmaPrec;
 T.Msg.SigmaMaxIt=Lev->SigmaMaxIt;
 AddHighPriorityTask(&T,sizeof(struct Task),Lev->TaskList);
}

static void CreateDirectSegTask(struct LevelCurve *Lev)
{
 struct Task T;
 struct Segment *Sd;

 Sd=GetFirstDataAdress(Lev->Orbit);
 sprintf(s,"D %d %ld %ld %ld %ld %d\n",
	 Lev->Id,
	 Sd->xi,Sd->yi,
	 Sd->xe,Sd->ye,(int) Sd->Flag);
 SendSock(SocketHandle,s);
 if (Lev->OrbitFlags & LIMIT_2_TRIANGLES)
   return;
 T.Tag=INTERVAL_TAG;
 T.Msg.z1.r=Sd->xi*Lev->Tau+Sd->yi*Lev->Tau*COSPI_3;
 T.Msg.z1.i=Sd->yi*Lev->Tau*SINPI_3;
 T.Msg.z2.r=Sd->xe*Lev->Tau+Sd->ye*Lev->Tau*COSPI_3;
 T.Msg.z2.i=Sd->ye*Lev->Tau*SINPI_3;
 T.Msg.SplitPrec=Lev->SplitPrec;
 T.Msg.SigmaPrec=Lev->SigmaPrec;
 T.Msg.SigmaMaxIt=Lev->SigmaMaxIt;
 T.Msg.Sigma=Lev->Sigma;
 T.Msg.Method=Lev->Method;
 T.Segment=Sd;
 AddLowPriorityTask(&T,sizeof(struct Task),Lev->TaskList);
}

static void CreateInverseSegTask(struct LevelCurve *Lev)
{
  struct Task T;
  struct Segment *Si;

  Si=GetLastDataAdress(Lev->Orbit);
  sprintf(s,"I %d %ld %ld %ld %ld %d\n",Lev->Id,
	  Si->xi,Si->yi,Si->xe,Si->ye,(int)Si->Flag);
  SendSock(SocketHandle,s);
  if (Lev->OrbitFlags & LIMIT_2_TRIANGLES)
    return;
  T.Tag=INTERVAL_TAG;
  T.Msg.z1.r=Si->xi*Lev->Tau+Si->yi*Lev->Tau*COSPI_3;
  T.Msg.z1.i=Si->yi*Lev->Tau*SINPI_3;
  T.Msg.z2.r=Si->xe*Lev->Tau+Si->ye*Lev->Tau*COSPI_3;
  T.Msg.z2.i=+Si->ye*Lev->Tau*SINPI_3;
  T.Msg.SplitPrec=Lev->SplitPrec;
  T.Msg.SigmaPrec=Lev->SigmaPrec;
  T.Msg.SigmaMaxIt=Lev->SigmaMaxIt;
  T.Msg.Sigma=Lev->Sigma;
  T.Msg.Method=Lev->Method;
  T.Segment=Si;
  AddLowPriorityTask(&T,sizeof(struct Task),Lev->TaskList);
}



int ProcessSlaveMessage(int Tag,struct SlaveMsg *Msg,struct ProcInfo *Proc)
{
  struct LevelCurve *Lev;
  struct Segment *Segment=(struct Segment *)Proc->Segment;
  struct Segment S,*S1;
  double f;

  /* Verifier si la ligne de niveau n'a pas ete annulee */
  Lev=LevelList;
  while(Lev!=NULL)
    {
      if (Lev==Proc->LevelCurve)
	break;
      Lev=Lev->Next;
    }
  if (Lev==NULL)
    return -1;

  switch (Tag)
   {
   case STARTUP_TAG:
     if (Msg->Status==0)
       {
	 sprintf(s,"X Could not start level curve\n");
	 SendSock(SocketHandle,s);
	 break;
       }
     S.xi= Msg->xi;
     S.yi= Msg->yi;
     S.xe= Msg->xe;
     S.ye= Msg->ye;
     f=Msg->ze.r;
     Add(S.xi,S.yi,Lev->Tau,&f,Lev->SigmaPrec);
     if (f>Lev->Sigma)
       {
	 sprintf(s,"X Could not start level curve\n");
	 SendSock(SocketHandle,s);
	 break;
       }
     f=Msg->ze.i;
     Add(S.xe,S.ye,Lev->Tau,&f,Lev->SigmaPrec);
     if (f<Lev->Sigma)
       {
	 sprintf(s,"X Could not start level curve\n");
	 SendSock(SocketHandle,s);
	 break;
       }
     S.Flag=Ze_CONV|Zi_CONV;
     if (Lev->OrbitFlags & ISREAL)
       {
	 if (S.yi<0)
	   {
	     S.xi=S.xi+S.yi;
	     S.yi=-S.yi;
	   }
	 if (S.ye<0)
	   {
	     S.xe=S.xe+S.ye;
	     S.ye=-S.ye;
	   }
       }
     AddFirst(&S,sizeof(S),Lev->Orbit);
     sprintf(s,"A %d %d %d %e %e %e %e %ld %.0f\n",
	     Lev->Id,Lev->OrbitFlags,Lev->Method,
	     Lev->Tau,Lev->Sigma,Lev->SigmaPrec,Lev->SplitPrec,
	     Lev->SigmaMaxIt,Lev->CreatedTime);
     SendSock(SocketHandle,s);    
     CreateDirectSegTask(Lev);
     PredictDirect(Lev);
     PredictInverse(Lev);
     break;
   case INTERVAL_TAG:
     Segment->z=Msg->z;
     Segment->Sigma=Msg->Sigma;
     Segment->Flag|=Zl_INCL;
     if (Msg->Status==1)
       Segment->Flag|=Zl_CONV;
     else
       Segment->Flag&=~Zl_CONV;
     Lev->LastModified=tic();
     sprintf(s,"L %ld %ld %ld %ld %ld %d %e %e %e %.0f\n",(long)Lev->Id,
	     Segment->xi,Segment->yi,Segment->xe,Segment->ye,
	     (int)Segment->Flag,
	     (double) Msg->z.r,(double) Msg->z.i,(double)Msg->Sigma,
	     (double) Lev->LastModified);
     SendSock(SocketHandle,s);
     break;
   case Z_TAG:
     S1=GetFirstDataAdress(Lev->Orbit);
     if (Msg->Status!=2)
       Add((long) S1->z.r,(long) S1->z.i,Lev->Tau,&Msg->Sigma,Lev->SigmaPrec);
     if (Msg->Sigma>Lev->Sigma)
       {
	 S1->xe=(long) S1->z.r;
	 S1->ye=(long) S1->z.i;
	 if (Msg->Status==1)
	   S1->Flag|=Ze_CONV;
	 else
	   S1->Flag &= ~Ze_CONV;
       }
     else
       {
	 S1->xi=(long) S1->z.r;
	 S1->yi=(long) S1->z.i;
	 if (Msg->Status==1)
	   S1->Flag|=Zi_CONV;
	 else
	   S1->Flag &= ~Zi_CONV;	 
       }
     CreateDirectSegTask(Lev);
     PredictDirect(Lev);
     break;
   case Z1_TAG:
     S1=GetLastDataAdress(Lev->Orbit);
     if (Msg->Status!=2)
       Add((long) S1->z.r,(long) S1->z.i,Lev->Tau,&Msg->Sigma,Lev->SigmaPrec);
     if (Msg->Sigma>Lev->Sigma)
       {
	 S1->xe=(long) S1->z.r;
	 S1->ye=(long) S1->z.i;
	 if (Msg->Status==1)
	   S1->Flag|=Ze_CONV;
 	 else
	   S1->Flag &= ~Ze_CONV;
      }
     else
       {
	 S1->xi=(long) S1->z.r;
	 S1->yi=(long) S1->z.i;
	 if (Msg->Status==1)
	   S1->Flag|=Zi_CONV;
 	 else
	   S1->Flag &= ~Zi_CONV;
      }
     CreateInverseSegTask(Lev);
     PredictInverse(Lev);
     break;
   }
  return 1;
}

int LevelID(void *L)
{
  struct LevelCurve *Lev=(struct LevelCurve *)L;
  if (Lev==NULL)
    return -1;
  else
    return Lev->Id;
}

int SetLevelCurveProc(int Id,int MinP,int MaxP)
{
  struct LevelCurve *L=LevelList;
  while (L!=NULL && L->Id!=Id)
    L=L->Next;
  if (L==NULL)
    return -1;
  L->MinP=MinP;
  L->MaxP=MaxP;
  return 1;
}

int SetLevelCurveAllProc(int Id)
{
  return SetLevelCurveProc(Id,-1,-1);
}

int SetLevelCurvePriority(int Id,int Priority)
{
  struct LevelCurve *L=LevelList,*p,*q;

  while(L!=NULL && L->Id!=Id)
    L=L->Next;
  if (L==NULL)
    return -1;
  L->Priority=Priority;
  L=NULL;
  while(LevelList)
    {
      p=L;
      while (p!=NULL && p->Priority>=LevelList->Priority)
	{
	  q=p;
	  p=p->Next;
	}
      if (p==L)
	{
	  L=LevelList;
	  LevelList=LevelList->Next;
	  L->Next=p;
	}
      else
	{
	  q->Next=LevelList;
	  LevelList=LevelList->Next;
	  q->Next->Next=p;
	}
    }
  LevelList=L;
  return 1;
}


void SetSocketHandle(int s)
{
  SocketHandle=s;
}


void LimitLevelCurve2Triangles(int Id,int v)
{
  struct LevelCurve *L=LevelList;
  while (L!=NULL && L->Id!=Id)
    L=L->Next;
  if (L==NULL)
    return;
  if (v)
    L->OrbitFlags = L->OrbitFlags | LIMIT_2_TRIANGLES;
  else
    L->OrbitFlags = L->OrbitFlags & (~LIMIT_2_TRIANGLES);
}
