#include "sparse.h"
#include "mmio.h"
#include <malloc.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <errno.h>
#include <sys/time.h>
#include <unistd.h>

#define COMPLEX_MATRIX   1
#define REAL_MATRIX      2

struct Sparse
{
  integer Lines,Columns,Flag;
  integer *Col;
  integer *Lin;
  Complex *Val;
};


/* Internal Function used to allocate structure */

static struct Sparse* SparseAllocate(long Row,long Col,long Nz)
{
  struct Sparse *R;
  
  if ((R=(struct Sparse *)malloc(sizeof(struct Sparse)))==NULL) 
    { 
      printf("Sparse Allocate: Out of memory (structure matrix)\n"); 
      exit(-1); 
    } 
  R->Lines=Row; 
  R->Columns=Col; 
  if ((R->Val=(Complex *)malloc(Nz*sizeof(Complex)))==NULL)
    { 
       printf("Sparse Allocate: Out of memory (%ld Complex)\n",Nz);
       exit(-1);
    } 
  if ((R->Lin=(long *)malloc(Nz*sizeof(long)))==NULL)
    { 
       printf("Sparse Allocate: Out of memory\n");
       exit(-1);
     }
   if ((R->Col=(long *)malloc((Col+1)*sizeof(long)))==NULL)
     {
       printf("Sparse Allocate: Out of memory\n");
       exit(-1);
     } 
   memset(R->Col,0,(Col+1)*sizeof(long)); 
  return R;
}

/* Function to load an mtx matrix */




void *SparseLoadMTX(char *filename)
{
 FILE *handle;
 char S[100];
 int l,c,nz,i,Empty,ColumnNb;
 struct Sparse *M;
 double f, g;
 MM_typecode matcode;
 
 if ((handle=fopen(filename,"rt"))==NULL)
   {
     perror("fopen");
     return NULL;
   }
 if (mm_read_banner(handle, &matcode) != 0)
    {
        printf("Could not process Matrix Market banner ");
        printf(" in file [%s]\n", filename);
        return NULL;
    }
 if (!(mm_is_matrix(matcode) && mm_is_sparse(matcode)))
   {
     printf("%s: does not define a sparse matrix\n",filename);
     return NULL;
   }
 
 if (mm_read_mtx_crd_size(handle, &l, &c, &nz) !=0)
    {
      printf("Could not parse matrix size.\n");
      return NULL;
    } 
 M=SparseAllocate(l,c,nz);
 if (mm_is_real(matcode))
   {
     M->Flag=REAL_MATRIX;
     fgets(S,100,handle);
     sscanf(S,"%d %d %lf",&l,&c,&f);
     M->Val[0].r=f;
     M->Val[0].i=0.0;
     M->Lin[0]=l-1;
     for(i=0;i<c;i++)
       M->Col[i]=0;
     M->Col[c]=1;
     Empty=1;
     ColumnNb=c;
     while(fgets(S,100,handle))
       {
	 sscanf(S,"%d %d %lf",&l,&c,&f);
	 M->Val[Empty].r=f;
	 M->Val[Empty].i=0.0;
	 M->Lin[Empty]=l-1;
	 if (c!=ColumnNb)
	   {
	     for(i=ColumnNb;i<c;i++)
	       M->Col[i]=Empty;
	     ColumnNb=c;
	   }	 
	 Empty++;
      }
     M->Col[c]=Empty;
   }
 if (mm_is_complex(matcode))
   {
     M->Flag=COMPLEX_MATRIX;
     fgets(S,100,handle);
     sscanf(S,"%d %d %lf %lf",&l,&c,&f,&g);
     M->Val[0].r=f;
     M->Val[0].i=g;
     M->Lin[0]=l-1;
     for(i=0;i<c;i++)
       M->Col[i]=0;
     M->Col[c]=1;
     Empty=1;
     ColumnNb=c;
     while(fgets(S,100,handle))
       {
	 if (S[0]=='%')
	   continue;
	 sscanf(S,"%d %d %lf %lf",&l,&c,&f,&g);
	 M->Val[Empty].r=f;
	 M->Val[Empty].i=g;
	 M->Lin[Empty]=l-1;
	 if (c!=ColumnNb)
	   {
	     for(i=ColumnNb;i<c;i++)
	       M->Col[i]=Empty;
	     ColumnNb=c;
	   }	 
	 Empty++;
      }
     M->Col[c]=Empty;
   }
 fclose(handle);
 return (void *)M;
}

/*  Function to destroy a Sparse Matrix  */

void SparseDestroy(void *ParA)
{
  struct Sparse *A=(struct Sparse *)ParA;
  if (A==NULL)
    return;
  free(A->Col);
  free(A->Val);
  free(A->Lin);
  free(A);
}



Complex SparseGet(void *parA,long l,long c,int *Status)
{
  Complex Tmp={0.0,0.0};
  long jmin,jmax,jmid;
  struct Sparse *A=(struct Sparse *)parA;
  
  if (A==NULL || A->Lines<l ||A->Columns<c|| l<0 || c<0)
    {
      printf("SparseGet: Error in parameters\n");
      *Status=0;
      return Tmp;
    }
  jmin=A->Col[c];
  jmax=A->Col[c+1];
  while(jmin<=jmax)
    {
      jmid=(jmin+jmax)>>1;
      if (A->Lin[jmid]==l)
	return A->Val[jmid];
      if (A->Lin[jmid]<l)
	jmin=jmid+1;
      else
	jmax=jmin-1;
    }
  *Status=1;
  return Tmp;
}

static void QuickSort(long *Ind1,long *Ind2,Complex *Val,long n)
{
  long i,j,tmp;
  Complex ctmp;
  if(n<2) return;
  if (Ind2)
    {
      i=0;
      n--;
      j=n-1;
      while(1)
	{
	  while (Ind1[i]<=Ind1[n] && i<n) i++;
	  while (Ind1[j]>=Ind1[n] && j>0) j--;
	  if (i<j)
	    {
	      tmp=Ind1[i];
	      Ind1[i]=Ind1[j];
	      Ind1[j]=tmp;
	      tmp=Ind2[i];
	      Ind2[i]=Ind2[j];
	      Ind2[j]=tmp;
	      ctmp=Val[i];
	      Val[i]=Val[j];
	      Val[j]=ctmp;
	    }
	  else
	    {
	      tmp=Ind1[i];
	      Ind1[i]=Ind1[n];
	      Ind1[n]=tmp;
	      tmp=Ind2[i];
	      Ind2[i]=Ind2[n];
	      Ind2[n]=tmp;
	      ctmp=Val[i];
	      Val[i]=Val[n];
	      Val[n]=ctmp;
	      QuickSort(Ind1,Ind2,Val,i);
	      QuickSort(Ind1+i+1,Ind2+i+1,Val+i+1,n-i);
	      break;
	    }
	}
    }
  else
    {
      i=0;
      n--;
      j=n-1;
      while(1)
	{
	  while (Ind1[i]<=Ind1[n] && i<n) i++;
	  while (Ind1[j]>=Ind1[n] && j>0) j--;
	  if (i<j)
	    {
	      tmp=Ind1[i];
	      Ind1[i]=Ind1[j];
	      Ind1[j]=tmp;
	      ctmp=Val[i];
	      Val[i]=Val[j];
	      Val[j]=ctmp;
	    }
	  else
	    {
	      tmp=Ind1[i];
	      Ind1[i]=Ind1[n];
	      Ind1[n]=tmp;
	      ctmp=Val[i];
	      Val[i]=Val[n];
	      Val[n]=ctmp;
	      QuickSort(Ind1,NULL,Val,i);
	      QuickSort(Ind1+i+1,NULL,Val+i+1,n-i);
	      break;
	    }
	}
    }
}

static void SparseToColFormat(struct Sparse *A)
{
  long *ALin,Nz,i,j;
  
  Nz=A->Lin[A->Lines];
  ALin=(long *)malloc(Nz*sizeof(long));
  if (ALin==NULL)
    {
      printf("SparseToColFormat: Out of memory\n");
      exit(-1);
    }
  for (i=0;i<A->Lines;i++)
    for(j=A->Lin[i];j<A->Lin[i+1];j++)
      ALin[j]=i;
  QuickSort(A->Col,ALin,A->Val,Nz);
  free(A->Lin);
  A->Lin=ALin;
  if ((ALin=(long *)malloc((A->Columns+1)*sizeof(long)))==NULL)
    {
      printf("SparseToColFormat: Out of memory\n");
      exit(-1);
    }
  memset(ALin,0,(A->Columns+1)*sizeof(long));
  for(i=0;i<Nz;i++)
    ALin[A->Col[i]+1]++;
  for(i=0;i<A->Columns;i++)
    ALin[i+1]+=ALin[i];
  free(A->Col);
  A->Col=ALin;
  for(i=0;i<A->Columns;i++)
    QuickSort(&A->Lin[A->Col[i]],NULL,&A->Val[A->Col[i]],A->Col[i+1]-A->Col[i]);

}

static int SparseToRowFormat(struct Sparse *A)
{
  long *ACol,Nz,i,j;
  
  Nz=A->Col[A->Columns];
  ACol=(long *)malloc(Nz*sizeof(long));
  if (ACol==NULL)
    {
      printf("SparseToRowFormat: Out of memory\n");
      exit(-1);
    }
  for (i=0;i<A->Columns;i++)
    for(j=A->Col[i];j<A->Col[i+1];j++)
      ACol[j]=i;
  QuickSort(A->Lin,ACol,A->Val,Nz);
  free(A->Col);
  A->Col=ACol;
  if ((ACol=(long *)malloc((A->Lines+1)*sizeof(long)))==NULL)
    {
      printf("SparseToRowFormat: Out of memory\n");
      exit(-1);
    }
  memset(ACol,0,(A->Lines+1)*sizeof(long));
  for(i=0;i<Nz;i++)
    ACol[A->Lin[i]+1]++;
  for(i=0;i<A->Lines;i++)
    ACol[i+1]+=ACol[i];
  free(A->Lin);
  A->Lin=ACol;
  for(i=0;i<A->Lines;i++)
    QuickSort(&A->Col[A->Lin[i]],NULL,&A->Val[A->Lin[i]],A->Lin[i+1]-A->Lin[i]);
  return 1;
}
/****************************************************/

void SparsePrintToFile(void *A,char *filename)
{
  struct Sparse *M=(struct Sparse *)A;
  FILE *f;
  long i,j;

  f=fopen(filename,"w");
  for(i=0;i<M->Columns;i++)
    for(j=M->Col[i];j<M->Col[i+1];j++)
      fprintf(f,"%d %d %e %e\n",M->Lin[j]+1,i+1,M->Val[j].r,M->Val[j].i);
  fclose(f);
  
}


/****************************************************/

/* Function to permute the columns of the matrix */

void SparsePermute(void *ParA,long *P)
{
  struct Sparse *A1,*A=(struct Sparse *)ParA;
  long l,i,j;

  if (P==NULL || A==NULL)
    return;
  A1=SparseAllocate(A->Lines,A->Columns,A->Col[A->Columns]);
  l=0;
  A1->Col[0]=0;
  for(i=0;i<A->Columns;i++)
    {
/*       j=A->Col[P[i]+1]-A->Col[P[i]]; */
/*       memcpy(&A1->Val[A1->Col[i]],&A->Val[A->Col[P[i]]],j*sizeof(Complex)); */
/*       memcpy(&A1->Lin[A1->Col[i]],&A->Lin[A->Col[P[i]]],j*sizeof(integer)); */
/*       A1->Col[i+1]=A1->Col[i]+j; */
      
      for(j=A->Col[P[i]];j<A->Col[P[i]+1];j++)
	{
	  A1->Val[l]=A->Val[j];
	  A1->Lin[l]=A->Lin[j];
	  l++;
	}
      A1->Col[i+1]=l;
    }
  free(A->Col);
  free(A->Lin);
  free(A->Val);
  A->Col=A1->Col;
  A->Val=A1->Val;
  A->Lin=A1->Lin;
  free(A1);
}


/* Prints the matrix in sparse format */

void SparsePrint(void *ParA)
{
  long i,j,jend;
  struct Sparse *M=(struct Sparse *)ParA;
  if (M==NULL)
    return;
  printf("Lines=%ld, Columns=%ld\n",M->Lines,M->Columns);
  printf("Type=%s\n",M->Flag==COMPLEX_MATRIX?"COMPLEX":
	 (M->Flag==REAL_MATRIX?"REAL":"UNKNOWN"));
  for(i=0;i<M->Columns;i++)
    {
      jend=M->Col[i+1];
      for(j=M->Col[i];j<jend;j++)
	printf("%ld %ld %e %e\n",M->Lin[j],i,M->Val[j].r,M->Val[j].i);
    }
}

/* Computer y=Ax, P is a permutation vector  P=NULL if no permutations are required*/

void SparseVect(void *ParA,char T, Complex *x, Complex *y,long *P) 
     /* Calcule y=A*x  si T='N' 
        Calcule y=A.'*x si T='T'
        Calcule y=A'*x si T='H'
     */
{
  struct Sparse *A=(struct Sparse *) ParA;

  if (A==NULL || x==NULL || y==NULL)
    {
      printf("SparseVect: Error in parameters\n");
      exit(-1);
    }
  if (P==NULL)
    switch (T)
      {
      case 'N':
      case 'n':
	{
	  long c,l,j,jend;
	  Complex alc,xc;
	  
	  memset(y,0,A->Columns*sizeof(Complex));
	  for(c=0;c<A->Columns;c++)
	    {
	      jend=A->Col[c+1];
	      for(j=A->Col[c];j<jend;j++)
		{
		  l=A->Lin[j];
		  alc=A->Val[j];
		  xc=x[c];
		  y[l].r+=alc.r*xc.r-alc.i*xc.i;
		  y[l].i+=alc.i*xc.r+alc.r*xc.i;
		}
	    }
	  break;
	}
      case 'T':
      case 't':
	{
	  long c,l,j,jend;
	  Complex alc,xl;

	  memset(y,0,A->Lines*sizeof(Complex));	
	  for(c=0;c<A->Columns;c++)
	    {
	      jend=A->Col[c+1];
	      for(j=A->Col[c];j<jend;j++)
		{
		  l=A->Lin[j];
		  alc=A->Val[j];
		  xl=x[l];
		  y[c].r+=alc.r*xl.r-alc.i*xl.i;
		  y[c].i+=alc.i*xl.r+alc.r*xl.i;
		}
	    }
	  break;
	}
      case 'H':
      case 'h':
	{
	  long c,j,jend;
	  Complex alc,xl;
	  
	  memset(y,0,A->Lines*sizeof(Complex));	
	  for(c=0;c<A->Columns;c++)
	    {
	      jend=A->Col[c+1];
	      for(j=A->Col[c];j<jend;j++)
		{
		  alc=A->Val[j];
		  xl=x[A->Lin[j]];
		  y[c].r+=alc.r*xl.r+alc.i*xl.i;
		  y[c].i+=-alc.i*xl.r+alc.r*xl.i;
		}
	    }
	  break;
	}
      }
  else
    switch (T)
      {
      case 'N':
      case 'n':
	{
	  long c,l,j,jend;
	  Complex alc,xc;
           
          memset(y,0,A->Columns*sizeof(Complex));
	  for(c=0;c<A->Columns;c++)
	    {
	      jend=A->Col[P[c]+1];
	      for(j=A->Col[P[c]];j<jend;j++)
		{
		  l=A->Lin[j];
		  alc=A->Val[j];
		  xc=x[c];
		  y[l].r+=alc.r*xc.r-alc.i*xc.i;
		  y[l].i+=alc.i*xc.r+alc.r*xc.i;
		}
	    }
	  break;
	}
      case 'T':
      case 't':
	{
	  long c,l,j,jend;
	  Complex alc,xl;
          
	  memset(y,0,A->Lines*sizeof(Complex));
	  for(c=0;c<A->Columns;c++)
	    {
	      jend=A->Col[P[c]+1];
	      for(j=A->Col[P[c]];j<jend;j++)
		{
		  l=A->Lin[j];
		  alc=A->Val[j];
		  xl=x[l];
		  y[c].r+=alc.r*xl.r-alc.i*xl.i;
		  y[c].i+=alc.i*xl.r+alc.r*xl.i;
		}
	    }
          break;
	}
      case 'H':
      case 'h':
	{
	  long c,l,j,jend;
	  Complex alc,xl;
	  
          memset(y,0,A->Lines*sizeof(Complex));
	  for(c=0;c<A->Columns;c++)
	    {
	      jend=A->Col[P[c]+1];
	      for(j=A->Col[P[c]];j<jend;j++)
		{
		  l=A->Lin[j];
		  alc=A->Val[j];
		  xl=x[l];
		  y[c].r+=alc.r*xl.r+alc.i*xl.i;
                   y[c].i+=-alc.i*xl.r+alc.r*xl.i;
		}
	    }
          break;
	}
      }
}


void *SparseAdd_zI(void *ParA,Complex b)
{
  struct Sparse *C;  /* C=A+b*B */
  long Nz,i,j,jend,h;
  struct Sparse *A=(struct Sparse *)ParA;

  if (A==NULL)
    {
      printf("SparseAdd: Error in parameters\n");
      return NULL;
    }
  if (b.r==0.0 && b.i==0)
    {
      C=SparseAllocate(A->Lines,A->Columns,A->Col[A->Columns]);
      memcpy(C->Col,A->Col,(A->Columns+1)*sizeof(long));
      memcpy(C->Lin,A->Lin,A->Col[A->Columns]*sizeof(long));
      memcpy(C->Val,A->Val,A->Col[A->Columns]*sizeof(Complex));
      return C;
    }
  Nz=A->Col[A->Columns]+A->Columns;
  C=SparseAllocate(A->Lines,A->Columns,Nz);
  h=0;
  C->Col[0]=0;
  for(i=0;i<A->Columns;i++)
    {
      jend=A->Col[i+1];
      j=A->Col[i];
      while(j<jend && A->Lin[j]<i)
	{
	  C->Lin[h]=A->Lin[j];
	  C->Val[h].r=A->Val[j].r;
	  C->Val[h].i=A->Val[j].i;
	  h++;
	  j++;
	}
      if (A->Lin[j]==i)
	{
	  C->Lin[h]=i;
	  C->Val[h].r=A->Val[j].r+b.r;
	  C->Val[h].i=A->Val[j].i+b.i;
	  h++;
	  j++;
	}
      else
	{
	  C->Lin[h]=i;
	  C->Val[h].r=b.r;
	  C->Val[h].i=b.i;
	  h++;
	}
      while(j<jend)
	{
	  C->Lin[h]=A->Lin[j];
	  C->Val[h].r=A->Val[j].r;
	  C->Val[h].i=A->Val[j].i;
	  h++;
	  j++;
	}
      C->Col[i+1]=h;
    }
  return (void *)C;
}

void MatProd(void *ParR,Complex *v,Complex *Av)
{
  long n,k,i,j;
  Complex xi,akk;
  double s;
  struct Sparse *R=(struct Sparse *)ParR;
  
  n=R->Lines;
  memcpy(Av,v+n,n*sizeof(Complex));
  for(k=n-1;k>=0;k--)
    {
      i=R->Col[k+1]-1;
      akk=R->Val[i];
      s=akk.r*akk.r+akk.i*akk.i;
      xi.r=(Av[k].r*akk.r+Av[k].i*akk.i)/s;
      xi.i=(-Av[k].r*akk.i+Av[k].i*akk.r)/s;
      Av[k]=xi;
      while(--i>=R->Col[k])
	{
	  j=R->Lin[i];
	  Av[j].r-=R->Val[i].r*xi.r-R->Val[i].i*xi.i;
	  Av[j].i-=R->Val[i].i*xi.r+R->Val[i].r*xi.i;
	}
    }
  Av+=n;
  memcpy(Av,v,n*sizeof(Complex));
  for(k=0;k<n;k++)
    {
      for(i=R->Col[k];i<R->Col[k+1]-1;i++)
	{
	  Av[k].r-=R->Val[i].r*Av[R->Lin[i]].r+R->Val[i].i*Av[R->Lin[i]].i;
	  Av[k].i-=R->Val[i].r*Av[R->Lin[i]].i-R->Val[i].i*Av[R->Lin[i]].r;
	}
      xi=R->Val[R->Col[k+1]-1];
      s=xi.r*xi.r+xi.i*xi.i;
      akk.r=(Av[k].r*xi.r-Av[k].i*xi.i)/s;
      akk.i=(Av[k].r*xi.i+Av[k].i*xi.r)/s;
      Av[k]=akk;
    }
}



long SparseColumns(void *A)
{
  return ((struct Sparse *)A)->Columns;
}

#ifdef MPI_CC

#include "mpi.h"
#include "messages.h"

void BroadCastMatrix(void *ParA,int Size)
{
  long Dims[3],i,k;
  MPI_Request *Req;
  MPI_Status *Stat;
  struct Sparse *A=(struct Sparse *)ParA;

  int blen[3];
  MPI_Aint indices[3];
  MPI_Datatype oldtypes[3];
  Complex z;
  MPI_Datatype MPIComplex,MPI_Dims,MPI_Col,MPI_Lin,MPI_Val;

  blen[0]=1; indices[0]=0;oldtypes[0]=MPI_DOUBLE;
  blen[1]=1; indices[1]=((void*)&z.i)-((void *)&z);oldtypes[1]=MPI_DOUBLE;
  blen[2]=1; indices[2]=sizeof(Complex);oldtypes[2]=MPI_UB;
  MPI_Type_struct(3,blen,indices,oldtypes,&MPIComplex);

  MPI_Type_vector(1,3,1,MPI_LONG,&MPI_Dims);
  MPI_Type_vector(1,A->Columns+1,1,MPI_LONG,&MPI_Col);
  MPI_Type_vector(1,A->Col[A->Columns],1,MPI_LONG,&MPI_Lin);
  MPI_Type_vector(1,A->Col[A->Columns],1,MPIComplex,&MPI_Val);
  Dims[0]=A->Lines;
  Dims[1]=A->Columns;
  Dims[2]=A->Col[A->Columns];
  Req=(MPI_Request *)malloc(4*Size*sizeof(MPI_Request));
  Stat=(MPI_Status *)malloc(4*Size*sizeof(MPI_Status));
  k=0;
  for(i=1;i<Size;i++)
    MPI_Isend(Dims,1,MPI_Dims,i,MATRIX_DIM_TAG,MPI_COMM_WORLD,&Req[k++]); 
  for(i=1;i<Size;i++)
    {
      MPI_Isend(A->Col,1,MPI_Col,i,A_COL_TAG,
		MPI_COMM_WORLD,&Req[k++]);
      MPI_Isend(A->Lin,1,MPI_Lin,i,A_LIN_TAG,
		MPI_COMM_WORLD,&Req[k++]);
      MPI_Isend(A->Val,1,MPI_Val,i,A_VAL_TAG,
		MPI_COMM_WORLD,&Req[k++]);
    }
  MPI_Waitall(k,Req,Stat);
  free(Req);
  free(Stat);
}


void *RecvMatrix()
{
  MPI_Request *Req;
  MPI_Status *Stat;
  struct Sparse *A;
  long Dims[3];

  int blen[3];
  MPI_Aint indices[3];
  MPI_Datatype oldtypes[3];
  Complex z;
  MPI_Datatype MPIComplex,MPI_Dims,MPI_Col,MPI_Lin,MPI_Val;


  blen[0]=1; indices[0]=0;oldtypes[0]=MPI_DOUBLE;
  blen[1]=1; indices[1]=((void*)&z.i)-((void *)&z);oldtypes[1]=MPI_DOUBLE;
  blen[2]=1; indices[2]=sizeof(Complex);oldtypes[2]=MPI_UB;
  MPI_Type_struct(3,blen,indices,oldtypes,&MPIComplex);

  Req=(MPI_Request *)malloc(9*sizeof(MPI_Request));
  Stat=(MPI_Status *)malloc(9*sizeof(MPI_Status));
  MPI_Type_vector(1,3,1,MPI_LONG,&MPI_Dims);
  MPI_Recv(Dims,1,MPI_Dims,0,MATRIX_DIM_TAG,MPI_COMM_WORLD,Stat);
  A=SparseAllocate(Dims[0],Dims[1],Dims[2]);

  MPI_Type_vector(1,Dims[1]+1,1,MPI_LONG,&MPI_Col);
  MPI_Type_vector(1,Dims[2],1,MPI_LONG,&MPI_Lin);
  MPI_Type_vector(1,Dims[2],1,MPIComplex,&MPI_Val);

  MPI_Irecv(A->Col,1,MPI_Col,0,A_COL_TAG,MPI_COMM_WORLD,&Req[0]);
  MPI_Irecv(A->Lin,1,MPI_Lin,0,A_LIN_TAG,MPI_COMM_WORLD,&Req[1]);
  MPI_Irecv(A->Val,1,MPI_Val,0,A_VAL_TAG,MPI_COMM_WORLD,&Req[2]);
  MPI_Waitall(3,Req,Stat);
  free(Req);
  free(Stat);
  return (void *)A;
}
  
#endif



struct List
{
  long Node;
  struct List *Next;
};

long EliminationTree(long *Srow,long *Col,long Lines,
		     long Columns,long *Pere,long *RSRow,
		     long *RowP,long *ColP)
     /* Computes the Elimination tree for the Sparse MultiFrontal QR decomposition */
{
  struct List **A,*p,*q;
  register long min,i,j;
  char *Test,Diag;
  register long NzR=0;
  /* Build intersection graph */
  A=(struct List **)malloc(Columns*sizeof(struct List *));
  Test=(char *)malloc(Columns*sizeof(char));
  for(i=0;i<Columns;i++)
    {
      A[i]=NULL;
      Test[i]=0;
    }
  /* No permutations */
  for(i=0;i<Lines;i++)
    {
      min=i;
      Diag=1;
      /* Compute the smallest column of nonzero entries in line i and 
	 check for the diagonal */
      for(j=Srow[i];j<Srow[i+1];j++)
	{
	  if (Col[j]<min)
	    min=Col[j];
	  if (Col[j]==i)
	    Diag=0;
	}
      /* Create the adjacency graph of row min */
      for(j=Srow[i];j<Srow[i+1];j++)
	{
	  p=(struct List *)malloc(sizeof(struct List));
	  p->Node=Col[j];
	  p->Next=A[min];
	  A[min]=p;
	}
      /* If non diagonal, assume one */
      if (Diag)
	{
	  p=(struct List *)malloc(sizeof(struct List));
	  p->Node=i;
	  p->Next=A[min];
	  A[min]=p;
	}
    }
  /* Compute fill ins */
  for(i=0;i<Columns-1;i++)
    {
      min=Columns-1;
      p=A[i];
      RSRow[i]=NzR;
      NzR++;
      while(p)
	{
	  if (Test[p->Node]==0 && p->Node>i)
	    {
	      Test[p->Node]=1;
	      NzR++;
	      if (p->Node<min)
		min=p->Node;
	    }
	  p=p->Next;
	}
      Pere[i]=min;
      p=A[i];
      while(p)
	{
	  if (Test[p->Node]==1)
	    {
	      Test[p->Node]=0;
	      if (p->Node!=min)
		{
		  q=p;
		  p=p->Next;
		  q->Next=A[min];
		  A[min]=q;
		}
	      else
		{
		  q=p;
		  p=p->Next;
		  free(q);
		}
	    }
	  else
	    {
	      q=p;
	      p=p->Next;
	      free(q);
	    }
	}
      A[i]=NULL;
    }
  RSRow[Columns-1]=NzR;
  NzR++;
  RSRow[Columns]=NzR;
  free(A);
  free(Test);
  return NzR;
}

struct FrontalMatrix
{
  long MaxLines,MaxColumns; /* Maximum dimension */
  long Lines,Columns;
  long *Lin,*Col;
  Complex *Val;
  struct FrontalMatrix *Next;
};



static struct FrontalMatrix *Trashed=NULL; /* keep unused frontal matrices for further use */

static void TrashFrontal(struct FrontalMatrix *Mat)
{
  if (Trashed==NULL)
    Trashed=Mat;
  else
    if (Trashed->MaxLines*Trashed->MaxColumns<Mat->MaxLines*Mat->MaxColumns)
      {
	free(Trashed->Lin);
	free(Trashed->Col);
	free(Trashed->Val);
	free(Trashed);
	Trashed=Mat;
      }
    else
      {
	free(Mat->Lin);
	free(Mat->Col);
	free(Mat->Val);
	free(Mat);
      }
}

static struct FrontalMatrix *BuildFrontalMatrix(long Lines,long Columns)
{
  struct FrontalMatrix *p;
  if (Trashed==NULL || Trashed->MaxLines<Lines || Trashed->MaxColumns<Columns)
    {
#ifdef DEBUG
      printf("Allocating %d x %d\n",Lines,Columns); 
#endif
      p=(struct FrontalMatrix *)malloc(sizeof(struct FrontalMatrix));
      p->MaxLines=p->Lines=Lines;
      p->MaxColumns=p->Columns=Columns;
      p->Lin=(long *)malloc(Lines*sizeof(long));
      p->Col=(long *)malloc(Columns*sizeof(long));
      p->Val=(Complex *)malloc(Lines*Columns*sizeof(Complex));
    }
  else
    {
      p=Trashed;
      Trashed=NULL;
#ifdef DEBUG
      printf("Reusing %d x %d\n",p->MaxLines,p->MaxColumns); 
#endif
      p->Lines=Lines;
      p->Columns=Columns;
    }
  return p;
}


static void DenseReduce(struct FrontalMatrix *A,Complex *Work)
{
  Complex Tau;
  long i,One=1;
  memcpy(Work,A->Val,A->Lines*sizeof(Complex));
  /* Compute Householder vector */
  zlarfg_(&A->Lines,&Work[0],&Work[1],&One,&Tau);
  Work[0].r=1.0;
  Work[0].i=0.0;
  Tau.i=-Tau.i;
  /* Apply Householder reflexion to dense matrix */
  zlarf_("L",&A->Lines,&A->Columns,&Work[0],&One,&Tau,A->Val,&A->Lines,&Work[A->Lines],(short)1);
}

static int ReduceNode(long CurrentNode,long FatherNode,
		      long *SRow,long *Col,Complex *Val,long Lines,long Columns,
		      long *RSRow,long *RCol,Complex *RVal,
		      struct FrontalMatrix **ContribFM,
		      long *CPattern,Complex *Work)
{
  struct FrontalMatrix *NodeFM,*CurrentContrib;
  long min,i,j,k,kend,l;
  long *p,*q,*stack;
  long *pi,*qi;
  long NodeFMLines=0,NodeFMColumns=0;
  
   /* CPattern must be created and set to 0 */
  /* Building Column Pattern of Frontal Matrix */
  stack=(long *)Work;
  stack[0]=1;
  for(j=0;j<Lines;j++)
    {
      /* We assume that the entries are not sorted!!! therefore we must look for the
	 first non zero entry */
      min=CurrentNode+1;
      for(i=SRow[j];i<SRow[j+1];i++)
	{
	  if (Col[i]<CurrentNode)
	    {
	      min=Col[i];
	      break;
	    }
	  if (Col[i]<min)
	    min=Col[i];
	}
      if (min==CurrentNode)
	{
	  stack[stack[0]++]=j;
	  NodeFMLines++;
	  p=&Col[SRow[j]];
	  q=&Col[SRow[j+1]];
	  while(p!=q)
	    {
	      if (CPattern[*p]>=0)
		{
		  NodeFMColumns++;
		  CPattern[*p]=-1;
		}
	      p++;
	    }
	}
    }
  /* Appending Pattern of contribution matrix */
  CurrentContrib=ContribFM[CurrentNode];
  while (CurrentContrib!=NULL)
    {
      NodeFMLines+=CurrentContrib->Lines-1; /* Remove first row from frontal matrix */
      pi=&CurrentContrib->Col[1]; /* Remove first column from frontal matrix */
      qi=&CurrentContrib->Col[CurrentContrib->Columns];
      while(pi!=qi)
	{
	  if (CPattern[*pi]>=0)
	    {
	      NodeFMColumns++;
	      CPattern[*pi]=-1;
	    }
	  pi++;
	}
      CurrentContrib=CurrentContrib->Next;	  
    }
  if (NodeFMLines==0)
    {
      memset(CPattern,0,Columns*sizeof(long));
      return 1;
    }
  NodeFM=BuildFrontalMatrix(NodeFMLines,NodeFMColumns);
  memset(NodeFM->Val,0,NodeFM->Lines*NodeFM->Columns*sizeof(Complex));
  k=0;
  for(j=0;j<Columns;j++)
    if (CPattern[j]==-1)
      {
	CPattern[j]=k;
	NodeFM->Col[k]=j;
	k++;
      }
  l=0;
  /* Remove the lines from the stack to compute the frontal matrix */
  for(i=1;i<stack[0];i++)
    {
      j=stack[i];
      kend=SRow[j+1];
      NodeFM->Lin[l]=j;
      k=SRow[j];
      p=&Col[k];
      q=&Col[SRow[j+1]];
      while (p!=q)
	NodeFM->Val[l+NodeFM->Lines*CPattern[*p++]]=Val[k++];
      l++;
    }
#ifdef DEBUG
  printf("CurrentNode=%d\n",CurrentNode);
  printf("   ");
  for(i=0;i<NodeFM->Columns;i++)
    printf("%2d ",NodeFM->Col[i]);
  for(i=0;i<NodeFM->Lines;i++)
    {
      printf("\n%2d  ",NodeFM->Lin[i]);
      for(j=0;j<NodeFM->Columns;j++)
	printf("%c  ",NodeFM->Val[i+j*NodeFM->Lines]!=0.0?'x':'.');
    }
  printf("\n");
#endif
  while(ContribFM[CurrentNode]!=NULL)
    {
      CurrentContrib=ContribFM[CurrentNode];
      ContribFM[CurrentNode]=CurrentContrib->Next;
      for(j=1;j<CurrentContrib->Lines;j++)
	{
	  NodeFM->Lin[l]=CurrentContrib->Lin[j];
	  for(k=1;k<CurrentContrib->Columns;k++)
	    {
	      NodeFM->Val[l+NodeFM->Lines*CPattern[CurrentContrib->Col[k]]]=
		CurrentContrib->Val[j+k*CurrentContrib->Lines];
	    }
	  l++;
	}
      TrashFrontal(CurrentContrib);
#ifdef DEBUG
      printf("Appending Contribution blocs\n",CurrentNode);
      printf("   ");
      for(i=0;i<NodeFM->Columns;i++)
	printf("%2d ",NodeFM->Col[i]);
      for(i=0;i<NodeFM->Lines;i++)
	{
	  printf("\n%2d  ",NodeFM->Lin[i]);
	  for(j=0;j<NodeFM->Columns;j++)
	    printf("%c  ",NodeFM->Val[i+j*NodeFM->Lines]!=0.0?'x':'.');
	}
      printf("\n");
      getchar();
#endif
    }
  DenseReduce(NodeFM,Work);
  if (RSRow[CurrentNode+1]-RSRow[CurrentNode]<NodeFM->Columns)
    {
      printf("Problem!!! Predicted size of row %d is %d; actual size is %d\n",
	     CurrentNode,RSRow[CurrentNode+1]-RSRow[CurrentNode],NodeFM->Columns);
      exit(-1);
    }
  memcpy(&RCol[RSRow[CurrentNode]],NodeFM->Col,NodeFM->Columns*sizeof(long));
  for(j=0;j<NodeFM->Columns;j++)
    RVal[RSRow[CurrentNode]+j]=NodeFM->Val[j*NodeFM->Lines];
  if (RSRow[CurrentNode+1]-RSRow[CurrentNode]>NodeFM->Columns)
    for(j=RSRow[CurrentNode]+NodeFM->Columns;j<RSRow[CurrentNode+1];j++)
      RCol[j]=-1;
  if (CurrentNode<Columns-1 && NodeFM->Lines>1 && NodeFM->Columns>1)
    {
      NodeFM->Next=ContribFM[FatherNode];
      ContribFM[FatherNode]=NodeFM;
    }
  else
    TrashFrontal(NodeFM);
  return 1;
} 

static int ReduceNodes(long *Pere,
		       long *SRow,long *Col,Complex *Val,long Lines,long Columns,
		       long *RSRow,long *RCol,Complex *RVal,
		       struct FrontalMatrix **ContribFM)
{
  long i,j,k;
  long *ChildCount;
  Complex *Work;
  long *CPattern;

  Work=(Complex *)malloc((Lines+Columns)*sizeof(Complex));
  CPattern=(long *)malloc(Columns*sizeof(long));
  memset(CPattern,0,Columns*sizeof(long));
  for(i=0;i<Columns;i++)
    ReduceNode(i,Pere[i],SRow,Col,Val,Lines,Columns,
	       RSRow,RCol,RVal,ContribFM,CPattern,Work);
  free(CPattern);
  free(Work);
}

int ComplexMFQR(long *SRow,long *Col,Complex *Val,long Lines,long Columns,
	       long **RSRow,long **RCol,Complex **RVal,long *RLines,long *RColumns)
{
  long *Pere,i,j,k,NzR;
  struct FrontalMatrix **ContribFM;
  struct timeval tv0,tv1;
  struct timezone tz0;
  
  gettimeofday(&tv0,&tz0);
  if (SRow==NULL || Col==NULL || Val==NULL || Lines<=0 || Columns <=0)
    return 0;
  /* Allocate necessary space for temporary storage */
  ContribFM=(struct FrontalMatrix **)
            malloc(Columns*sizeof(struct FrontalMatrix *));
  if (ContribFM==NULL)
    return 0;
  memset(ContribFM,0,Columns*sizeof(struct FrontalMatrix *));
  /* Allocate space for elimination tree */
  Pere=(long *)malloc(Columns*sizeof(long));
  /* Allocate space for the pattern of R */
  *RSRow=(long *)malloc((Columns+1)*sizeof(long));
  /* Build Elimination tree and the pattern of R 
     On return NzR is the number of nonzero entries in R */
  NzR=EliminationTree(SRow,Col,Lines,Columns,Pere,*RSRow,NULL,NULL);
  /* Allocate space for the R matrix */
  *RCol=(long *)malloc(NzR*sizeof(long));
  *RVal=(Complex *)malloc(NzR*sizeof(Complex));
  ReduceNodes(Pere,SRow,Col,Val,Lines,Columns,
	      *RSRow,*RCol,*RVal,ContribFM);
  free(ContribFM);
  free(Pere);
  if (Trashed)
    {
      free(Trashed->Lin);
      free(Trashed->Col);
      free(Trashed->Val);
      free(Trashed);
      Trashed=NULL;
    }
  *RLines=Lines<Columns?Lines:Columns;
  *RColumns=Columns;
  /* Suppress useless fill-in */
  k=0;
  j= (*RSRow)[0];
  for(i=0;i<*RLines;i++)
    {
      for(;j< (*RSRow)[i+1];j++)
	if ((*RCol)[j]!=-1)
	  {
	    (*RCol)[k]=(*RCol)[j];
	    (*RVal)[k]=(*RVal)[j];
	    k++;
	  }
      (*RSRow)[i+1]=k;
    }
#ifdef DEBUG
  gettimeofday(&tv1,&tz0);
  printf("Time=%fs\n",tv1.tv_sec-tv0.tv_sec+1e-6*(tv1.tv_usec-tv0.tv_usec));
#endif
  return 1;
}

void *SparseMFQR(void *ParA)
{
  struct Sparse *R;
  struct Sparse *A=(struct Sparse *)ParA;
  if (A==NULL)
    {
      printf("SparseMFQR: NULL matrix\n");
      return NULL;
    }
  SparseToRowFormat(A);
  R=(struct Sparse *)malloc(sizeof(struct Sparse));
  
  ComplexMFQR(A->Lin,A->Col,A->Val,A->Lines,A->Columns,
	      &R->Lin,&R->Col,&R->Val,&R->Lines,&R->Columns);

SparseToColFormat(R);
  return (void *)R;
}
